#include <jss/dataflow.hpp>
#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <algorithm>
#include <atomic>
#include <ncurses.h>

unsigned const width=50;
unsigned const height=30;

jss::dataflow::broadcast_channel<bool> cells[width][height];
jss::dataflow::broadcast_channel<bool> heartbeat;

bool cell_rules(std::vector<bool> const& incoming){
  bool const was_alive=incoming[0];
  unsigned const alive_neighbours=
    std::count(incoming.begin()+1,incoming.end()-1,true);
  return (was_alive && (alive_neighbours==2)) || 
    (alive_neighbours==3);
}

void add_column(std::vector<jss::dataflow::readable_channel<bool> >& vec,unsigned x,unsigned y){
  vec.push_back(cells[x][y]);
  if(y>0){
    vec.push_back(cells[x][y-1]);
  }
  if(y<(height-1)){
    vec.push_back(cells[x][y+1]);
  }
}

std::vector<jss::dataflow::readable_channel<bool> > find_neighbours(unsigned x,unsigned y){
  std::vector<jss::dataflow::readable_channel<bool> > vec;
  vec.reserve(10);
  add_column(vec,x,y);
  if(x>0){
    add_column(vec,x-1,y);
  }
  if(x<(width-1)){
    add_column(vec,x+1,y);
  }
  return vec;
}

void bind_cell_evolution_rules(){
  for(unsigned x=0;x<width;++x){
    for(unsigned y=0;y<height;++y){
      std::vector<jss::dataflow::readable_channel<bool> > vec=
        find_neighbours(x,y);
      vec.push_back(heartbeat);
      jss::dataflow::combine(vec).
        transform(cell_rules).write_to(cells[x][y]);
    }
  }
}

void set_initial_grid(){
  for(unsigned x=0;x<width;++x){
    for(unsigned y=0;y<height;++y){
        cells[x][y]<<(rand()%2);
    }
  }
}

void update_display(std::vector<bool> const& cell_data){
  for(unsigned x=0;x<width;++x){
    for(unsigned y=0;y<height;++y){
      mvaddch(y,x,cell_data[y*width+x]?'*':' ');
    }
  }
  refresh();
}

void set_output(){
  std::vector<jss::dataflow::readable_channel<bool> > vec;
  for(unsigned y=0;y<height;++y){
    for(unsigned x=0;x<width;++x){
      vec.push_back(cells[x][y]);
    }
  }
  jss::dataflow::combine(vec).chain(update_display);
}

int main(){
  initscr();
  cbreak();
  curs_set(0);
  noecho();
  timeout(1000);
  bind_cell_evolution_rules();
  set_output();
  set_initial_grid();
  for(;;){
    int ch=getch();
    if(ch!=ERR){
      if(ch=='p'){
        mvaddstr(height,0,"Paused");
        timeout(-1);
        do{
          ch=getch();
        } while((ch!='p') && (ch!='q'));
        timeout(1000);
        mvaddstr(height,0,"      ");
      }
      if(ch=='q'){
        break;
      }
    }
    heartbeat<<true;
  }
  endwin();
}
